﻿// Copyright (C) Microsoft. All rights reserved. 
﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Runtime.InteropServices; // Need for Filzbach
using Microsoft.Research.Science.Data; // Need for Scientific DataSet
using Microsoft.Research.Science.Data.Imperative; // Need for Scientific DataSet
using Microsoft.Research.Science.Data.Climate.Conventions; // Need for FetchClimate

using FilzbachInterop; // Need for Filzbach

// This is an example of a model that you could write to fit the parameters to 
// using this solution. If you want to do your own model then you need to write 
// a new public class e.g. MyNewModel.cs which follows the conventions of the "ModelBase" abstract class
// See BlankClass.cs for an example. 

namespace CSToolsCodeTemplate
{
    /// <summary>
    /// A class of functions to enable the assessment of the likelihood of a particular model given the data. Provides the
    /// algorithms needed to make the model predictions.
    /// In this case the algorithms implement a model of net primary productivity (NPP) from 
    /// Lieth, H, 1972: Modelling the primary productivity of the world. 
    /// Nature and Resources, UNESCO, VIII, 2:5-10.
    /// 
    /// The model has been generalised to allow key parameters to be inferred
    /// 
    /// The model predicts the net carbon fixation (Net Primary Productivity) per year.
    /// The default parameters assume units of kg C m-2 y-1
    /// 
    /// To construct an instance of the model use MiamiNPPModel
    /// 
    /// This model reads in data on plant net primary productivity rates from the Ecosystem Modelling Data Intercomparison
    /// project,
    /// Olson, R. J., J. M. O. Scurlock, S. D. Prince, D. L. Zheng, and K. R. Johnson (eds.). 2001. 
    /// NPP Multi-Biome: NPP and Driver Data for Ecosystem Model-Data Intercomparison. Data set. 
    /// Available on-line [http://www.daac.ornl.gov] from the Oak Ridge National Laboratory 
    /// Distributed Active Archive Center, Oak Ridge, Tennessee, U.S.A. 
    /// (For details of this dataset see EMDINPPData.txt)
    public class MIAMINPP : ModelBase
    {
        public MIAMINPP() : base("MIAMINPP") { }

        /// <summary>
        /// Assumed maximum NPP (kg C m-2 y-1)
        /// </summary>
        /// <remarks>
        /// Model parameters that could be inferred must be marked with the FilzbachParameter attribute. 
        /// This atribute data is used to initialize the property value (Val) and to initialize Filzbach parameter properties.
        /// </remarks>
        [FilzbachParameter(lb: 0.6, ub: 3, val: 2.3, type: 1, @fixed: 0, dsply: 1)]
        public double max_NPP { get; set; }

        /// <summary>
        /// Parameter controlling the NPP response to precipitation
        /// </summary>
        [FilzbachParameter(0.000001, 0.01, 0.000664, 1, 0, 1)]
        public double p_NPP { get; set; }

        /// <summary>
        /// Parameter controlling the NPP response to temperature
        /// </summary>
        [FilzbachParameter(0.01, 1, 0.33, 1, 0, 1)]
        public double t1_NPP { get; set; }

        /// <summary>
        /// Parameter controlling the NPP response to temperature
        /// </summary>
        [FilzbachParameter(0.01, 1, 0.119, 1, 0, 1)]
        public double t2_NPP { get; set; }

        /// <summary>
        /// The "noise" parameter
        /// </summary>
        [FilzbachParameter(0.000001, 100, 1, 1, 0, 1)]
        public double th0_NPP { get; set; }


        /// <summary>
        /// The function used to make the model predictions to be compared with observations and assess the likelihood. In this case
        /// the function implements the MIAMI model cited above.
        /// </summary>
        /// <param name="SiteData">The dataset used to provide environmental data</param>
        /// <returns>A double array of predictions</returns>
        public override double EstimatePredictionsAndLikelihood(DataSet SiteData)
        {
            double[] RecordNumber = SiteData.GetData<double[]>("RecordNumber");
            int NumberOfRecords = RecordNumber.Length; // get the number of records
            if (!SiteData.Variables.Contains("MAT")) GetMATFetchClimate(SiteData); // If mean annual temperature (MAT) is not in the dataset then put it there using Fetchclimate
            if (!SiteData.Variables.Contains("MAP")) GetMAPFetchClimate(SiteData); // If mean annual precipitation (MAP) is not in the dataset then put it there using Fetchclimate

            double[] MAT = SiteData.GetData<double[]>("MAT"); //get MAT from the dataset
            double[] MAP = SiteData.GetData<double[]>("MAP"); // get MAP from the dataset

            double[] prediction = new double[NumberOfRecords]; // container to hold predictions for each site
            double NPP_T = 0.0;
            double NPP_P = 0.0;
            for (int ii = 0; ii < NumberOfRecords; ii++)
            {
                NPP_T = max_NPP / (1 + Math.Exp(t1_NPP - t2_NPP * MAT[ii])); // make predictions using the model for each site
                NPP_P = max_NPP * (1 - Math.Exp(-p_NPP * MAP[ii]));
                prediction[ii] = Math.Min(NPP_T, NPP_P);
            }
            SiteData.PutData<double[]>("Predictions", prediction);

            double likelihood = 0.0;
            var dataName = "Data";
            if (SiteData.Variables.Contains(dataName) && SiteData[dataName].TypeOfData == typeof(double) && SiteData[dataName].Rank == 1)
            {
                double[] data = SiteData.GetData<double[]>(dataName); //get the mean annual temperature

                for (int ii = 0; ii < NumberOfRecords; ii++)
                {
                    likelihood += Math.Log(Filzbach.normal_density(data[ii], prediction[ii], th0_NPP));
                }
            }

            return likelihood;
        }

        /// <summary>
        /// Uses FetchClimate to get mean annual temperature (degrees celcius) for each location in the dataset (specified by a list of latitude and longitudes)
        /// and put it into the dataset. 
        /// </summary>
        /// <param name="SiteData">The dataset containing the site data</param>
        private static void GetMATFetchClimate(DataSet SiteData)
        {
            using (DataSet fc = GetYearMeanRequest(SiteData))
            {
                fc.Fetch(ClimateParameter.FC_TEMPERATURE, "airt", dataSource: EnvironmentalDataSource.CRU_CL_2_0); //this call will create 2D variable on dimensions records and months and fill it with a FetchClimate 
                double[] air = fc.GetData<double[]>("airt", DataSet.ReduceDim(0), DataSet.FromToEnd(0));

                for (int i = 0; i < air.Length; i++) if (double.IsNaN(air[i])) air[i] = 0;
                SiteData.Add<double[]>("MAT", air, "RecordNumber");
                if (!SiteData.Variables.Contains("Predictions")) SiteData.Add<double[]>("Predictions", air,"RecordNumber");
            }
        }

        /// <summary>
        /// Uses FetchClimate to get mean annual precipitation (mm) for each location in the dataset (specified by a list of latitude and longitudes)
        /// and put it into the dataset. 
        /// </summary>
        /// <param name="SiteData">The dataset containing the site data</param>
        private static void GetMAPFetchClimate(DataSet SiteData)
        {
            using (DataSet fc = GetYearMeanRequest(SiteData))
            {
                fc.Fetch(ClimateParameter.FC_PRECIPITATION, "prate", dataSource: EnvironmentalDataSource.CRU_CL_2_0); //this call will create 2D variable on dimensions records and months and fill it with a FetchClimate 
                double[] prate = fc.GetData<double[]>("prate", DataSet.ReduceDim(0), DataSet.FromToEnd(0));
                for (int i = 0; i < prate.Length; i++)
                {
                    if (double.IsNaN(prate[i]))
                    {
                        prate[i] = 0;
                    }
                    else
                    {
                        prate[i] *= 12.0;
                    }
                }
                SiteData.Add<double[]>("MAP", prate, "RecordNumber");
                if (!SiteData.Variables.Contains("Predictions")) SiteData.Add<double[]>("Predictions", prate, "RecordNumber");
            }

        }

        /// <summary>
        /// Used to identify the dataset to use
        /// </summary>
        /// <param name="fileName">the name of the file... not used in this case</param>
        /// <returns></returns>
        public override DataSet MakeData(string fileName)
        {
            DataSet SourceData = DataSet.Open("MIAMINPPFC.csv"); //Use if you want to use FetchClimate to get the data
            //DataSet SourceData = DataSet.Open("MIAMINPP.csv"); // Use if you don't want to use FetchClimate
            return SourceData;
        }

        /// <summary>
        /// Creates and returns new memory dataset prepared to perform Fetch(), where 
        /// spatial regions are taken from the InputData, and
        /// temporal settings is 1 interval "30 years from 1961 to 1990".
        /// </summary>
        /// <param name="InputData">The input data containing the latitude and longitudes</param>
        /// <returns>A memory dataset</returns>
        static DataSet GetYearMeanRequest(DataSet InputData)
        {
            var ds = DataSet.Open("msds:memory2");

            //adding climatology bounds variable          
            ds.AddClimatologyAxisYearly(yearmin: 1961, yearmax: 1990, yearStep: 30);

            ds.AddVariableByValue(InputData.Variables["Latitude"]); //copying Latitude and Longitude variables into new dataset
            ds.AddVariableByValue(InputData.Variables["Longitude"]);

            return ds; //now this dataset is ready for data fetching from FetchClimate
        }
    }
}
