﻿// Copyright (C) Microsoft. All rights reserved. 
﻿using System;
using System.Collections.Generic;
using System.Reflection;
using Microsoft.Research.Science.Data;

// This code was written to generalise the 
// process of declaring a model for fitting
// within this solution. You shouldn't need to 
// change any of this code.

namespace CSToolsCodeTemplate
{
    public abstract class ModelBase
    {
        public string Name { get; private set; }
        private Dictionary<string, Tuple<PropertyInfo, FilzbachParameterAttribute>> fp = null;

        public ModelBase(string name)
        {
            Name = name;
            fp = new Dictionary<string, Tuple<PropertyInfo, FilzbachParameterAttribute>>();
            foreach (var p in this.GetType().GetProperties())
            {
                var atts = p.GetCustomAttributes(typeof(FilzbachParameterAttribute), false);
                if (atts.Length > 0)
                {
                    var att = (FilzbachParameterAttribute)atts[0];
                    if (p.PropertyType == typeof(double) || p.PropertyType == typeof(double[]))
                    {
                        if (p.PropertyType.IsArray && att.Number <= 0)
                            throw new ApplicationException(
                            string.Format("Property {0}.{1} has invalid attribute FilzbachParameterAttribute.Number {2}. For vector properties Number must be >0.",
                            p.DeclaringType.FullName, p.Name, att.Number));
                        fp[p.Name] = Tuple.Create(p, att);
                        p.GetSetMethod().Invoke(this, new object[] { att.Val });
                    }
                    else
                    {
                        throw new ApplicationException(
                            string.Format("Property {0}.{1} has invalid type {2}. Supported types are double and double[]",
                            p.DeclaringType.FullName, p.Name, p.PropertyType.FullName));
                    }
                }
            }
        }

        public void SetFilzbachProperty(string name, object value)
        {
            fp[name].Item1.GetSetMethod().Invoke(this, new object[] { value });
        }

        public abstract DataSet MakeData(string dataSetName);
        public abstract double EstimatePredictionsAndLikelihood(DataSet SiteData);

        public Dictionary<string, FilzbachParameterAttribute> FilzbachParameters
        {
            get
            {
                var result = new Dictionary<string, FilzbachParameterAttribute>();
                foreach (var kv in fp) result[kv.Key] = kv.Value.Item2;
                return result;
            }
        }
    }

    /// <summary>
    /// Marks a property as a Filzbach parameter.
    /// </summary>
    [AttributeUsage(AttributeTargets.Property)]
    public class FilzbachParameterAttribute : Attribute
    {
        public double Lb { get; set; }
        public double Ub { get; set; }
        public double Val { get; set; }
        public int Type { get; set; }
        public int Fixed { get; set; }
        public int Dsply { get; set; }
        public int Number { get; set; }
        /// <summary>
        /// Mark a property as a vector Filzbach parameter.
        /// </summary>
        /// <param name="lb">Lower bound. Must be less than upper bound.</param>
        /// <param name="ub">Upper bound. Must be less than lower bound.</param>
        /// <param name="val">Initial parameter value.</param>
        /// <param name="type">Parameter type. 0 - normal scale (location parameter); 1 - logarithm scale (scale parameter).</param>
        /// <param name="fixed">Parameter fix. 0 - do not fix; 1 - fix at initial value.</param>
        /// <param name="dsply">Display parameter in Filzbach console output. 0 - do not include in output; 1 - include in output.</param>
        /// <param name="number">Number of elements in the vector.</param>
        public FilzbachParameterAttribute(double lb, double ub, double val, int type, int @fixed, int dsply, int number)
        {
            Lb = lb; Ub = ub; Val = val; Type = type; Fixed = @fixed; Dsply = dsply; Number = number;
        }
        /// <summary>
        /// Mark a property as a scalar Filzbach parameter.
        /// </summary>
        /// <param name="lb">Lower bound. Must be less than upper bound.</param>
        /// <param name="ub">Upper bound. Must be less than lower bound.</param>
        /// <param name="val">Initial parameter value.</param>
        /// <param name="type">Parameter type. 0 - normal scale (location parameter); 1 - logarithm scale (scale parameter).</param>
        /// <param name="fixed">Parameter fix. 0 - do not fix; 1 - fix at initial value.</param>
        /// <param name="dsply">Display parameter in Filzbach console output. 0 - do not include in output; 1 - include in output.</param>
        public FilzbachParameterAttribute(double lb, double ub, double val, int type, int @fixed, int dsply)
            : this(lb, ub, val, type, @fixed, dsply, 0) { }
    }

}
