﻿// Copyright (C) Microsoft. All rights reserved. 
﻿using Microsoft.Research.Science.Data;
using FilzbachInterop;

// This code was written to generalise the 
// process of declaring a model for fitting
// within this solution. You shouldn't need to 
// change any of this code.

namespace CSToolsCodeTemplate
{
    public class ModelForInference
    {
        /// <summary>
        /// The name of the model
        /// </summary>
        public ModelBase Model;

        /// <summary>
        /// Makes a model for inference
        /// </summary>
        /// <param name="model">A model class conforming to ModelBase</param>
        public ModelForInference(ModelBase model)
        {
            this.Model = model;
        }

        /// <summary>
        /// Loads the parameters of the model into Filzbach
        /// </summary>
        public void SetupParameters()
        {
            foreach (var kv in Model.FilzbachParameters)
            {
                if (kv.Value.Number > 0)
                    Filzbach.parameter_create_vector(kv.Key, kv.Value.Lb, kv.Value.Ub, kv.Value.Val, kv.Value.Type, kv.Value.Fixed, kv.Value.Dsply, kv.Value.Number);
                else
                    Filzbach.parameter_create(kv.Key, kv.Value.Lb, kv.Value.Ub, kv.Value.Val, kv.Value.Type, kv.Value.Fixed, kv.Value.Dsply);

            }
        }

        /// <summary>
        /// Makes a prediction using the designated prediction equation
        /// </summary>
        /// <param name="observations">A dataset containting site data used by the prediction equation</param>
        /// <returns>A double array of predictions</returns>
        public double Likelihood(DataSet observations)
        {
            foreach (var kv in Model.FilzbachParameters)
            {
                var len = kv.Value.Number;
                if (len > 0)
                {
                    var v = new double[len];
                    for (int i = 0; i < len; i++) v[i] = Filzbach.parameter_getvalue(kv.Key, i);
                    Model.SetFilzbachProperty(kv.Key, v);
                }
                else
                {
                    var v = Filzbach.parameter_getvalue(kv.Key);
                    Model.SetFilzbachProperty(kv.Key, v);
                }
            }
            return Model.EstimatePredictionsAndLikelihood(observations);
        }
    }
}
