﻿// Copyright (C) Microsoft. All rights reserved. 
﻿using System;
using Microsoft.Research.Science.Data;
using Microsoft.Research.Science.Data.Imperative;

using FilzbachInterop;

namespace CSToolsCodeTemplate
{
    /// <summary>
    /// A class of functions to enable the assessment of the likelihood of a particular model given the data. Provides the
    /// algorithms needed to make the model predictions.
    /// In this case the algorithms implement a linear relationship with noise
    /// </summary>
    public class MyModel : ModelBase
    {
        public MyModel() : base("MyModel") { }

        /// <summary>
        /// "Gradient" model parameter.
        /// </summary>
        /// <remarks>
        /// Model parameters that could be inferred must be marked with the FilzbachParameter attribute. 
        /// This atribute data is used to initialize the property value (Val) and to initialize Filzbach parameter properties.
        /// </remarks>
        [FilzbachParameter(lb: 0.6, ub: 3, val: 2.3, type: 1, @fixed: 0, dsply: 1)]
        public double gradient { get; set; }

        /// <summary>
        /// "intecept" model parameter
        /// </summary>
        [FilzbachParameter(0.11, 50, 12.4, 1, 0, 1)]
        public double intercept { get; set; }

        /// <summary>
        /// "noise" model parameter
        /// </summary>
        [FilzbachParameter(0.01, 4, 2.0, 1, 0, 1)]
        public double theta { get; set; }
        
        /// <summary>
        /// The function used to make the model predictions to be compared with observations and assess the likelihood
        /// </summary>
        /// <param name="SiteData">The dataset used to provide environmental data</param>
        /// <returns>A double array of predictions</returns>
        public override double EstimatePredictionsAndLikelihood(DataSet SiteData)
        {
            int NumberOfRecords = SiteData.Dimensions["RecordNumber"].Length; // get the number of records
            double[] Environment = SiteData.GetData<double[]>("x"); //get the mean annual temperature
            double[] prediction = new double[NumberOfRecords]; // containor to hold predictions for each site
            for (int ii = 0; ii < NumberOfRecords; ii++)
            {
                prediction[ii] = gradient * Environment[ii] + intercept; // make predictions usign the model for each site
            }
            SiteData.PutData<double[]>("prediction", prediction);

            double likelihood = 0.0;
            var dataName = "data";
            if (SiteData.Variables.Contains(dataName) && SiteData[dataName].TypeOfData == typeof(double) && SiteData[dataName].Rank == 1)
            {
                double[] data = SiteData.GetData<double[]>("data"); //get the mean annual temperature

                for (int ii = 0; ii < NumberOfRecords; ii++)
                {
                    likelihood += Math.Log(Filzbach.normal_density(data[ii], prediction[ii], theta));
                }

            }
            return likelihood;
        }

        /// <summary>
        /// Makes a .csv file containing faked data
        /// generated by the model we wish to fit with 
        /// specific parameters that we aim to infer.
        /// </summary>
        /// <param name="fileName">The name of the file</param>
        /// <returns>a dataset containing the faked data</returns>
        public override DataSet MakeData(string fileName)
        {
            string extension = ".csv";

            Console.WriteLine("Creating DataSet named " + fileName + extension);

            DataSet myData = DataSet.Open(fileName + extension + "?openMode=create");
           
            int numData = 100;

            int[] recordNumber = new int[numData];
            double[] x = new double[numData];
            double[] prediction = new double[numData];

            for (int ii = 0; ii < numData; ii++)
            {
                recordNumber[ii] = ii;
                x[ii] = (double)ii;
                prediction[ii] = 2.3 * x[ii] + 12.4;
                prediction[ii] = Filzbach.normal_draw(prediction[ii], 2.0);
            }

            myData.Add<int[]>("RecordNumber", recordNumber, "RecordNumber");
            myData.Add<double[]>("x", x, "RecordNumber");
            myData.Add<double[]>("prediction", prediction, "RecordNumber");
            myData.Add<double[]>("data", prediction, "RecordNumber");
            myData.Dispose();

            return myData;
        }
    }
}
