﻿// Copyright (C) Microsoft. All rights reserved. 
//
// Microsoft Computational Science Laboratory C# Tools 1.0 template
// 
// This is a code template containing examples for using a variety of tools produced
// by the Computational Science Group at Microsoft Research Cambridge, UK
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

/// Non standard libraries - in this case Scientific DataSet libraries that
/// allow us to read in and output datasets in various formats (mostly .csv files 
/// or .nc files)
using Microsoft.Research.Science.Data; // Used to enable Scientific Dataset commands
using Microsoft.Research.Science.Data.CSV; // Used to enable Scientific Dataset commands
using Microsoft.Research.Science.Data.Imperative; // Used to enable Scientific Dataset commands
using FilzbachInterop; // Used to enable Filzbach

namespace CSToolsCodeTemplate
{
    class Program
    {
        /// <summary>
        /// This is an individual holder for an individual model for which we 
        /// want to conduct inference on. You don't need to change this.
        /// </summary>
        static ModelForInference myModel;

        /// <summary>
        /// This is an empty DataSet object that will hold a
        /// memory clone of the original datafile to speed
        /// internal computations such as parameter inference.
        /// You don't need to change this.
        /// </summary>
        static DataSet DataSetClone;

        /// <summary>
        /// This is the main function that implements Filzbach inference on the parameters
        /// of a user created model, and then outputs the results to a data file. You will probably
        /// only need to change one line of this.
        /// </summary>
        /// <param name="args">Not Used: arguments passed to the main function, e.g. from the command line</param>
        static void Main(string[] args)
        {
            // Make a new model, in this case MyModel however the commented out lines
            // enable this to be done for another model. THen clone the dataset for that model.
            // This demo works on the principle that the user (you) defines the model in one
            // seperate class and you only need to identify this model here and where the source data
            // resides and you are away. You need to change this bit of code if you're going to 
            // work with your own model. Only implement one of the "var model..." lines below and leave
            // the other two commented out... or comment them all out and include your own class
            // conforming to the template (look at MyModel, BlankClass and MIAMINPP to see what you need).
            //var model = new MyModel(); // Option 1: Fitting a linear regression model to faked data
            //var model = new BlankClass(); // Option 2: Simply a container for a user specified model
            var model = new MIAMINPP(); // Option 3: Fit a model of annual carbon fixation rate to real data

            // We clone the dataset we want to fit the model to so that it resides in RAM rather than
            // on the computers hard disk - this makes the inference faster. You probably don't need to change this.
            CloneDataSet(model.Name, model.MakeData(model.Name));

            // Tell the user that we got here. You don't need to change this.
            Console.WriteLine("Adding model for inference called " + model.Name + ".");
            myModel = new ModelForInference(model);

            InitialiseFilzbach(Settings1.Default.AnalysisName); // You probably don't need to change this.

            // Tell the user that we got here. You probably don't need to change this.
            Console.WriteLine("Model successfully prepared for inference. Outputs will be labelled " + Settings1.Default.AnalysisName + ". Press any key to begin.");

            // Wait for a key press. You probably don't need to change this.
            Console.ReadKey();

            // Run the Markov Chain. You probably don't need to change this.
            Filzbach.runmcmc(Settings1.Default.BurnIn, Settings1.Default.MarkovChainLength, Settings1.Default.MaxLikeBurnIn, Settings1.Default.MaxLikeChainLength); // the main command to run MCMC model fitting

            // Output the results. You probably don't need to change this.
            OutputData();
        }

        /// <summary>
        /// Clones a physical dataset to a memory dataset. You probably don't need to change this.
        /// </summary>
        /// <param name="modelName">The name of the model</param>
        /// <param name="dataForInference">The physical dataset we wish to clone</param>
        static void CloneDataSet(string modelName, DataSet dataForInference)
        {
            Console.WriteLine("Using dataset with URI " + dataForInference.URI + " as the source for inferring the parameters of " + modelName + ".");
            DataSetUri MemoryDatasetURI = DataSetUri.Create("msds:memory2");
            DataSetClone = dataForInference.Clone(MemoryDatasetURI);
            dataForInference.Dispose();
        }

        /// <summary>
        /// Initialises Filzbach including
        /// 1. Clear memory
        /// 2. Initialise the parameters
        /// 3. Set the number of chains (only works for one chain right now)
        /// 4. Name the analysis
        /// You probably don't need to change this.
        /// </summary>
        /// <param name="analysisName">The name of the analysis</param>
        static void InitialiseFilzbach(string analysisName)
        {
            Console.WriteLine("Initialising Filzbach and setting up parameters for model named " +myModel.Model.Name);
            
            Filzbach.initialize_filzbach();
        
            myModel.SetupParameters();

            Console.WriteLine("Setting the number of Markov Chains as " + Settings1.Default.nChains.ToString() + ".");
            Filzbach.set_chains(Settings1.Default.nChains); // currently needed to avoid bugs

            Filzbach.name_analysis(Settings1.Default.AnalysisName);
        }

        /// <summary>
        /// Identifies the likelihood function. This should be 
        /// "public override double EstimatePredictionsAndLikelihood(DataSet SiteData)"
        /// in the model code that you write. See examples such as MyModel.cs.
        /// You don't need to change this.
        /// </summary>
        static void LikelihoodFunction()
        {
            Filzbach.set_metr_ltotnew(myModel.Likelihood(DataSetClone));
        }

        /// <summary>
        /// Assigns the function "Likelihood function" to be that used to calculate the 
        /// model likelihood during Filzbach.
        /// You don't need to change this.
        /// </summary>
        static Program()
        {
            Filzbach.pfn_likelihood = LikelihoodFunction;
        }

        /// <summary>
        /// Outputs the data to a NetCDF file which you can view in DataSet Viewer.
        /// You may want to change this to get different outputs or change the 
        /// output datafile type.
        /// </summary>
        static void OutputData()
        {
            // Select one of the 2 below
            DataSet OutputDataFile = DataSet.Open("msds:nc?file=" + Settings1.Default.AnalysisName + ".nc&openMode=create"); // Output a netcdf file
            ///DataSet OutputDataFile = DataSet.Open("msds:csv?file=" + Settings1.Default.AnalysisName + ".csv&openMode=create"); // Output a csv file
            OutputDataFile.IsAutocommitEnabled = false; // Set this to be false and then commit all the changes to the file at the end.

            int samples = 0; while (Filzbach.params_from_bayes_list(0,samples) == 0) samples+=1; // Add up the number of samples in the Bayes list
            int[] chainID = new int[samples]; // Index for the chain
            for (int ii = 0; ii < samples; ii++) chainID[ii] = ii;

            // Output results for each parameter
            foreach (string parameter in Filzbach.parameter_names)
            {
                double l95 = Filzbach.l95(parameter); // Output the lower 95% confidence interval of the posterior parameter distribution
                double u95 = Filzbach.u95(parameter); // Output the upper 95% confidence interval of the posterior parameter distribution
                double mean = Filzbach.postmean(parameter); // Output the mean of the posterior parameter distribution
                double med = Filzbach.postmedian(parameter); // Output the median of the posterior parameter distribution

                double[] chain = new double[samples]; // Record the samples from the Markov Chain 
                for (int ii = 0; ii < samples; ii++) 
                {
                    int r = Filzbach.params_from_bayes_list(0,ii);
                    chain[ii] = Filzbach.parameter_getvalue(parameter);
                }

                // Save in a datafile
                OutputDataFile.Add<double[]>(parameter + "_chain", chain, "chainID");
                OutputDataFile.Add<double>(parameter + "_l95", l95);
                OutputDataFile.Add<double>(parameter + "_u95", u95);
                OutputDataFile.Add<double>(parameter + "_med", med);
                OutputDataFile.Add<double>(parameter + "_mean", mean);
            }

            OutputDataFile.Add<int[]>("chainID", chainID, "chainID");
            OutputDataFile.Commit();
            OutputDataFile.Dispose();           
        }
    }
}
